@@ -33,6 +33,7 @@
 #include <glibtop/proctime.h>
 #include <glibtop/procuid.h>
 #include <glibtop/procargs.h>
+#include <glibtop/prockernel.h>
 #include <glibtop/mem.h>
 #include <glibtop/swap.h>
 #include <sys/stat.h>
@@ -42,8 +43,6 @@
 #include <set>
 #include <list>
 
-#include <libgnomevfs/gnome-vfs-utils.h>
-
 #include "procman.h"
 #include "selection.h"
 #include "proctable.h"
@@ -53,6 +52,8 @@
 #include "interface.h"
 #include "selinux.h"
 
+
+ProcInfo::UserMap ProcInfo::users;
 ProcInfo::List ProcInfo::all;
 std::map<pid_t, guint64> ProcInfo::cpu_times;
 
@@ -222,12 +223,11 @@ proctable_new (ProcData * const procdata)
 		N_("Security Context"),
 		N_("Command Line"),
 		N_("Memory"),
+		N_("wchan"),
 		NULL,
 		"POINTER"
 	};
 
-	g_assert(COL_MEM == 15);
-
 	gint i;
 
 	scrolled = gtk_scrolled_window_new (NULL, NULL);
@@ -252,6 +252,7 @@ proctable_new (ProcData * const procdata)
 				    G_TYPE_STRING,	/* Security Context */
 				    G_TYPE_STRING,	/* Arguments	*/
 				    G_TYPE_ULONG,	/* Memory       */
+				    G_TYPE_STRING,	/* wchan	*/
 				    GDK_TYPE_PIXBUF,	/* Icon		*/
 				    G_TYPE_POINTER,	/* ProcInfo	*/
 				    G_TYPE_STRING	/* Sexy tooltip */
@@ -294,7 +295,7 @@ proctable_new (ProcData * const procdata)
 	gtk_tree_view_set_expander_column (GTK_TREE_VIEW (proctree), column);
 
 
-	for (i = COL_USER; i <= COL_MEM; i++) {
+	for (i = COL_USER; i <= COL_WCHAN; i++) {
 
 		GtkCellRenderer *cell;
 		GtkTreeViewColumn *col;
@@ -445,28 +446,34 @@ get_process_name (ProcInfo *info,
 
 
 
-static void
-get_process_user(ProcData* procdata, ProcInfo* info, uid_t uid)
+void
+ProcInfo::set_user(guint uid)
 {
-	struct passwd* pwd;
-	char* username;
-
-	if(G_LIKELY(info->uid == uid))
+	if (G_LIKELY(this->uid == uid))
 		return;
 
-	info->uid = uid;
+	this->uid = uid;
 
-	pwd = getpwuid(uid);
+	typedef std::pair<ProcInfo::UserMap::iterator, bool> Pair;
+	ProcInfo::UserMap::value_type hint(uid, "");
+	Pair p(ProcInfo::users.insert(hint));
 
-	if(pwd && pwd->pw_name)
-		username = g_strdup(pwd->pw_name);
-	else
-		username = g_strdup_printf("%u", (unsigned)uid);
+	// procman_debug("User lookup for uid %u: %s", uid, (p.second ? "MISS" : "HIT"));
 
-	/* don't free, because info->user belongs to procdata->users */
-	info->user = g_string_chunk_insert_const(procdata->users, username);
+	if (p.second) {
+		char* username;
+		struct passwd* pwd;
+		pwd = getpwuid(uid);
 
-	g_free(username);
+		if (pwd && pwd->pw_name)
+			username = g_strdup(pwd->pw_name);
+		else
+			username = g_strdup_printf("%u", uid);
+
+		p.first->second = username;
+	}
+
+	this->user = p.first->second;
 }
 
 
@@ -541,6 +548,7 @@ update_info_mutable_cols(ProcInfo *info)
 	tree_store_update(model, &info->node, COL_START_TIME, info->start_time);
 	tree_store_update(model, &info->node, COL_NICE, info->nice);
 	tree_store_update(model, &info->node, COL_MEM, info->mem);
+	tree_store_update(model, &info->node, COL_WCHAN, info->wchan);
 }
 
 
@@ -633,6 +641,10 @@ update_info (ProcData *procdata, ProcInfo *info)
 	glibtop_proc_state procstate;
 	glibtop_proc_uid procuid;
 	glibtop_proc_time proctime;
+	glibtop_proc_kernel prockernel;
+
+	glibtop_get_proc_kernel(&prockernel, info->pid);
+	g_strlcpy(info->wchan, prockernel.wchan, sizeof info->wchan);
 
 	glibtop_get_proc_state (&procstate, info->pid);
 	info->status = procstate.state;
@@ -642,7 +654,7 @@ update_info (ProcData *procdata, ProcInfo *info)
 
 	get_process_memory_info(info);
 
-	get_process_user(procdata, info, procstate.uid);
+	info->set_user(procstate.uid);
 
 	info->pcpu = (proctime.rtime - info->cpu_time) * 100 / procdata->cpu_total_time;
 	info->pcpu = MIN(info->pcpu, 100);
@@ -677,11 +689,13 @@ ProcInfo::ProcInfo(pid_t pid)
 	/* FIXME : wrong. name and arguments may change with exec* */
 	get_process_name (info, procstate.cmd, arguments[0]);
 
-	char* tooltip = g_strjoinv(" ", arguments);
-	info->tooltip = g_markup_escape_text(tooltip, -1);
+	std::string tooltip = make_string(g_strjoinv(" ", arguments));
+	if (tooltip.empty())
+	  tooltip = procstate.cmd;
 
-	info->arguments = g_strescape(tooltip, "\\\"");
-	g_free(tooltip);
+	info->tooltip = g_markup_escape_text(tooltip.c_str(), -1);
+
+	info->arguments = g_strescape(tooltip.c_str(), "\\\"");
 	g_strfreev(arguments);
 
 	guint64 cpu_time = proctime.rtime;
@@ -938,6 +952,6 @@ ProcInfo::set_icon(Glib::RefPtr<Gdk::Pixbuf> icon)
   GtkTreeModel *model;
   model = gtk_tree_view_get_model(GTK_TREE_VIEW(ProcData::get_instance()->tree));
   gtk_tree_store_set(GTK_TREE_STORE(model), &this->node,
-		     COL_PIXBUF, this->pixbuf->gobj(),
+		     COL_PIXBUF, (this->pixbuf ? this->pixbuf->gobj() : NULL),
 		     -1);
 }
